/*******************************************************************************
/
/	File:			MPEGPlayer.cpp
/
/	Description:	Simple MPEG player for BeOS.
/
/	Copyright (C) 2000 Carlos Hasan. All rights reserved.
/
*******************************************************************************/

#include <Application.h>
#include <DirectWindow.h>
#include <SoundPlayer.h>
#include <MediaDefs.h>
#include <Locker.h>
#include <View.h>
#include <Menu.h>
#include <MenuItem.h>
#include <MenuBar.h>
#include <FilePanel.h>
#include <Alert.h>
#include <TextView.h>
#include <Screen.h>
#include <Entry.h>
#include <Path.h>
#include <Debug.h>
#include <libmpeg3.h>

////////////////////////////////////////////////////////////////////////////////

class AudioPlayer {
public:
	AudioPlayer(mpeg3_t *file, int track);
	
	virtual ~AudioPlayer();

public:
	bigtime_t CurrentTime() const;
	
	bigtime_t PerformanceTime() const;
	
	bigtime_t Latency() const;

	media_raw_audio_format Format() const;
	
	float Volume() const;
	
	void SetVolume(float volume);
	
	bool Repeat() const;
	
	void SetRepeat(bool repeat);
	
	status_t Start();
	
	void Stop();
	
	bool HasData() const;

public:
	long CountFrames() const;
	
	long CurrentFrame() const;
	
	bigtime_t Duration() const;
	
	void SeekToFrame(long frame);
	
	void SeekToTime(bigtime_t time);
	
	void ReadFrames(void *outFrames, int frameCount);
	
private:
	static void PlayBuffer(void *cookie, void *buffer, size_t size,
						   const media_raw_audio_format & format);
	
private:
	BSoundPlayer *fPlayer;
	mpeg3_t	*fFile;
	int fTrack;
	bool fRepeat;
};

AudioPlayer::AudioPlayer(mpeg3_t *file, int track)
	:	fPlayer(NULL),
		fFile(file),
		fTrack(track),
		fRepeat(false)
{
	PRINT(("AudioPlayer::AudioPlayer()\n"));
	
	media_raw_audio_format format = media_raw_audio_format::wildcard;
	
	format.frame_rate = mpeg3_sample_rate(fFile, fTrack);
	format.channel_count = mpeg3_audio_channels(fFile, fTrack);
	format.format = media_raw_audio_format::B_AUDIO_FLOAT;
	format.byte_order = B_MEDIA_HOST_ENDIAN;
	format.buffer_size = (int) ((0.1f * format.frame_rate * (format.format & 0x0f)) + 1023) & ~1023;
	
	fPlayer = new BSoundPlayer(&format, "MPEG Audio Player", PlayBuffer, NULL, this);
}

AudioPlayer::~AudioPlayer()
{
	PRINT(("AudioPlayer::~AudioPlayer()\n"));

	fPlayer->SetHasData(false);
	fPlayer->Stop();
	delete fPlayer;
}

bigtime_t AudioPlayer::CurrentTime() const
{
	return fPlayer->CurrentTime();
}

bigtime_t AudioPlayer::PerformanceTime() const
{
	return fPlayer->PerformanceTime();
}

bigtime_t AudioPlayer::Latency() const
{
	return fPlayer->Latency();
}

media_raw_audio_format AudioPlayer::Format() const
{
	return fPlayer->Format();
}

float AudioPlayer::Volume() const
{
	return fPlayer->VolumeDB();
}

bool AudioPlayer::HasData() const
{
	return fPlayer->HasData();
}

bool AudioPlayer::Repeat() const
{
	return fRepeat;
}

void AudioPlayer::SetRepeat(bool repeat)
{
	fRepeat = repeat;
}

void AudioPlayer::SetVolume(float volume)
{
	fPlayer->SetVolumeDB(volume);
}

status_t AudioPlayer::Start()
{
	fPlayer->SetHasData(true);
	return fPlayer->Start();
}

void AudioPlayer::Stop()
{
	fPlayer->Stop(false, true);
}

long AudioPlayer::CountFrames() const
{
	return mpeg3_audio_samples(fFile, fTrack);
}

long AudioPlayer::CurrentFrame() const
{
	return mpeg3_get_sample(fFile, fTrack);
}

bigtime_t AudioPlayer::Duration() const
{
	bigtime_t total_frames = mpeg3_audio_samples(fFile, fTrack);
	int frame_rate = mpeg3_sample_rate(fFile, fTrack);
	return (1000000LL * total_frames) / frame_rate;
}

void AudioPlayer::SeekToFrame(long frame)
{
	mpeg3_set_sample(fFile, frame, fTrack);
}

void AudioPlayer::SeekToTime(bigtime_t time)
{
	int frame_rate = mpeg3_sample_rate(fFile, fTrack);
	long frame = (time * frame_rate) / 1000000;
	mpeg3_set_sample(fFile, frame, fTrack);
}

void AudioPlayer::ReadFrames(void *outFrames, int frameCount)
{
	mpeg3_read_audio(fFile, (float *) outFrames, NULL, -1, frameCount, fTrack);
	
	if (fRepeat && CurrentFrame() >= CountFrames())
		SeekToFrame(0);
}

void AudioPlayer::PlayBuffer(void *cookie, void *buffer, size_t size,
						   		 const media_raw_audio_format & format)
{
	if (format.format == media_raw_audio_format::B_AUDIO_FLOAT) {
		int frameSize = sizeof(float) * format.channel_count;
		((AudioPlayer *) cookie)->ReadFrames(buffer, size / frameSize);
	}
}

////////////////////////////////////////////////////////////////////////////////

struct blit_rect {
	int src_left, src_top;
	int src_width, src_height;
	int dst_left, dst_top;
	int dst_width, dst_height;
};

class VideoPlayer {
public:
	VideoPlayer(mpeg3_t *file, int track);
	
	virtual ~VideoPlayer();

public:
	bigtime_t CurrentTime();

	bigtime_t PerformanceTime();
	
	bigtime_t Latency();

	media_raw_video_format Format();
	
	int Width();
	
	int Height();

	status_t Start();
	
	void Stop();

	bool Repeat() const;
	
	void SetRepeat(bool repeat);
	
	void SetDrawBuffer(void (*drawBuffer)(void *), void *cookie);
	
public:
	long CountFrames();
	
	long CurrentFrame();
	
	bigtime_t Duration();
	
	void SeekToFrame(long frame);
	
	void SeekToTime(bigtime_t time);

	void ReadFrame(unsigned char **outFrame, color_space space,
				   const blit_rect *rects, int count);

	void DropFrames(int frames);
	
private:
	void Run();

	static int32 DrawBuffer(void *cookie);

private:
	BLocker fLocker;
	mpeg3_t	*fFile;
	int fTrack;
	bool fRepeat;
	volatile bigtime_t fCurrentTime;
	bigtime_t fLatency;
	thread_id fThread;
	volatile bool fConnected;
	volatile void (*fDrawBuffer)(void *);
	void *fCookie;
};


VideoPlayer::VideoPlayer(mpeg3_t *file, int track)
	:	fLocker("Video Player Locker", true),
		fFile(file),
		fTrack(track),
		fRepeat(false),
		fCurrentTime(0),
		fLatency(0),
		fThread(0),
		fConnected(true),
		fDrawBuffer(NULL),
		fCookie(NULL)
{
	PRINT(("VideoPlayer::VideoPlayer()\n"));

	fThread = spawn_thread(DrawBuffer, "MPEG Video Player",
		B_DISPLAY_PRIORITY, this);
}

VideoPlayer::~VideoPlayer()
{
	PRINT(("VideoPlayer::~VideoPlayer()\n"));

	status_t status;

	fConnected = false;
	resume_thread(fThread);
	wait_for_thread(fThread, &status);
}

bigtime_t VideoPlayer::CurrentTime()
{
	return fCurrentTime;
}

bigtime_t VideoPlayer::PerformanceTime()
{
	long frame = mpeg3_get_frame(fFile, fTrack);
	float frame_rate = mpeg3_frame_rate(fFile, fTrack);
	return (bigtime_t) (1000000LL * (frame / frame_rate));
}

bigtime_t VideoPlayer::Latency()
{
	return fLatency;
}

bool VideoPlayer::Repeat() const
{
	return fRepeat;
}

void VideoPlayer::SetRepeat(bool repeat)
{
	fRepeat = repeat;
}

media_raw_video_format VideoPlayer::Format()
{
	media_raw_video_format format = media_raw_video_format::wildcard;

	format.field_rate = mpeg3_frame_rate(fFile, fTrack);
	format.interlace = 1;
	format.first_active = 0;
	format.last_active = mpeg3_video_height(fFile, fTrack) - 1;
	format.orientation = B_VIDEO_TOP_LEFT_RIGHT;
	format.display.format = B_RGB32;
	format.display.line_width = mpeg3_video_width(fFile, fTrack);
	format.display.line_count = mpeg3_video_height(fFile, fTrack);
	format.display.bytes_per_row = 4 * format.display.line_width;
	format.display.pixel_offset = 0;
	format.display.line_offset = 0;

	return format;
}

int VideoPlayer::Width()
{
	return mpeg3_video_width(fFile, fTrack);
}

int VideoPlayer::Height()
{
	return mpeg3_video_height(fFile, fTrack);
}

status_t VideoPlayer::Start()
{
	return resume_thread(fThread);
}

void VideoPlayer::Stop()
{
	suspend_thread(fThread);
}

long VideoPlayer::CountFrames()
{
	return mpeg3_video_frames(fFile, fTrack);
}

long VideoPlayer::CurrentFrame()
{
	return mpeg3_get_frame(fFile, fTrack);
}

bigtime_t VideoPlayer::Duration()
{
	bigtime_t total_frames = mpeg3_video_frames(fFile, fTrack);
	float frame_rate = mpeg3_frame_rate(fFile, fTrack);
	return (bigtime_t) (1000000LL * (total_frames / frame_rate));
}

void VideoPlayer::SeekToFrame(long frame)
{
	mpeg3_set_frame(fFile, frame, fTrack);
}

void VideoPlayer::SeekToTime(bigtime_t time)
{
	float frame_rate = mpeg3_frame_rate(fFile, fTrack);
	long frame = (long) ((time * frame_rate) / 1000000);
	mpeg3_set_frame(fFile, frame, fTrack);
}

void VideoPlayer::ReadFrame(unsigned char **outFrame, color_space space,
							const blit_rect *rect, int count)
{
	bigtime_t time = system_time();

	for (int index = 0; index < count; index++, rect++) {
		if (index == 0)
			mpeg3_read_frame(fFile, outFrame,
				rect->src_left, rect->src_top,
				rect->src_width, rect->src_height,
				rect->dst_width, rect->dst_height,
				space == B_RGB32 ? MPEG3_BGRA8888 :
				space == B_RGB24 ? MPEG3_BGR888 :
				space == B_RGB16 ? MPEG3_RGB565 :
				space == B_RGB32_BIG ? MPEG3_RGBA8888 :
				space == B_RGB24_BIG ? MPEG3_RGB888 : MPEG3_RGB565,
				fTrack);
		else
			mpeg3_reread_frame(fFile, outFrame,
				rect->src_left, rect->src_top,
				rect->src_width, rect->src_height,
				rect->dst_width, rect->dst_height,
				space == B_RGB32 ? MPEG3_BGRA8888 :
				space == B_RGB24 ? MPEG3_BGR888 :
				space == B_RGB16 ? MPEG3_RGB565 :
				space == B_RGB32_BIG ? MPEG3_RGBA8888 :
				space == B_RGB24_BIG ? MPEG3_RGB888 : MPEG3_RGB565,
				fTrack);
		
		outFrame += rect->dst_height;
	}
	
	fLatency = (fLatency + system_time() - time) >> 1;
	
	if (fRepeat && CurrentFrame() >= CountFrames())
		SeekToFrame(0);
}

void VideoPlayer::DropFrames(int frames)
{
	mpeg3_drop_frames(fFile, frames, fTrack);
}

void VideoPlayer::SetDrawBuffer(void (*drawBuffer)(void *), void *cookie)
{
	if (fLocker.Lock()) {
		fDrawBuffer = drawBuffer;
		fCookie = cookie;
		fLocker.Unlock();
	}
}

void VideoPlayer::Run()
{
	bigtime_t startTime = system_time() + 10000LL;
	float frameRate = mpeg3_frame_rate(fFile, fTrack);
	
	for (long frame = 1; fConnected; frame++) {
		fCurrentTime = startTime + (bigtime_t) ((1000000.0 * frame) / frameRate);
		if (fCurrentTime >= system_time()) {
			if (fLocker.Lock()) {
				if (fDrawBuffer != NULL)
					fDrawBuffer(fCookie);
				fLocker.Unlock();
			}
		}
		else {
			mpeg3_drop_frames(fFile, 1, fTrack);
			PRINT(("drop\n"));
		}
		snooze_until(fCurrentTime, B_SYSTEM_TIMEBASE);
	}
}

int32 VideoPlayer::DrawBuffer(void *cookie)
{
	((VideoPlayer *) cookie)->Run();
	return B_OK;
}

////////////////////////////////////////////////////////////////////////////////

enum {
	kOpen,
	kScaleHalf,
	kScaleNormal,
	kScaleDouble,
	kFullScreen,
	kLoop,
	kTiming,
	kPlay,
	kStop,
	kRewind,
	kForward,
	kBack
};

class VideoWindow : public BDirectWindow {
public:
	VideoWindow(BRect frame, const char *title, window_type type, uint32 flags);
	
	virtual ~VideoWindow();

	virtual void MessageReceived(BMessage *message);
	
	virtual void DirectConnected(direct_buffer_info *info);

	void OpenFile(entry_ref *ref);

	void CloseFile();

	void Start();
	
	void Stop();

	void SetScale(float scale);

	bool Repeat();
	
	void SetRepeat(bool repeat);
	
	bool Full();
	
	void SetFull(bool full);
	
	bigtime_t CurrentTime();
	
	void SeekToTime(bigtime_t time);
	
	void Draw();

	void CreateBuffer(direct_buffer_info *info, int width, int height, float menuBarHeight);
	
	void DeleteBuffer();
	
private:
	static void DrawBuffer(void *cookie);

private:
	BLocker fLocker;
	float fMenuBarHeight;

	BMenuBar *fMenuBar;
	BView *fView;
	mpeg3_t *fFile;
	VideoPlayer *fVideoPlayer;
	AudioPlayer *fAudioPlayer;
	
	bool fDirty;
	bool fConnected;
	direct_buffer_info *fInfo;
	
	color_space fBufferPixelFormat;
	int fBufferNumClips;
	unsigned char **fBufferRows;
	blit_rect *fBufferClips;
};

VideoWindow::VideoWindow(BRect frame, const char *title, window_type type, uint32 flags)
	:	BDirectWindow(frame, title, type, flags),
		fLocker(),
		fMenuBarHeight(0),
		fMenuBar(NULL),
		fView(NULL),
		fFile(NULL),
		fVideoPlayer(NULL),
		fAudioPlayer(NULL),
		fDirty(false),
		fConnected(false),
		fInfo(NULL),
		fBufferPixelFormat(B_NO_COLOR_SPACE),
		fBufferNumClips(0),
		fBufferRows(NULL),
		fBufferClips(NULL)
{
	PRINT(("VideoWindow::VideoWindow()\n"));
	
	BRect bounds(Bounds());
	
	fMenuBar = new BMenuBar(bounds, "MenuBar");
	BMenu *fileMenu = new BMenu("File");
	fileMenu->AddItem(new BMenuItem("Open...", new BMessage(kOpen), 'O', B_COMMAND_KEY));
	fileMenu->AddSeparatorItem();
	fileMenu->AddItem(new BMenuItem("About", new BMessage(B_ABOUT_REQUESTED)));
	fileMenu->AddSeparatorItem();
	fileMenu->AddItem(new BMenuItem("Close", new BMessage(B_QUIT_REQUESTED), 'W', B_COMMAND_KEY));
	fileMenu->AddItem(new BMenuItem("Quit", new BMessage(B_QUIT_REQUESTED), 'Q', B_COMMAND_KEY));
	fMenuBar->AddItem(fileMenu);
	
	BMenu *viewMenu = new BMenu("View");
	viewMenu->AddItem(new BMenuItem("50% scale", new BMessage(kScaleHalf), '0', B_COMMAND_KEY));
	viewMenu->AddItem(new BMenuItem("100% scale", new BMessage(kScaleNormal), '1', B_COMMAND_KEY));
	viewMenu->AddItem(new BMenuItem("200% scale", new BMessage(kScaleDouble), '2', B_COMMAND_KEY));
	viewMenu->AddItem(new BMenuItem("Full screen", new BMessage(kFullScreen), 'F', B_COMMAND_KEY));
	fMenuBar->AddItem(viewMenu);
	
	BMenu *controlMenu = new BMenu("Control");
	controlMenu->AddItem(new BMenuItem("Play", new BMessage(kPlay), 'P', B_COMMAND_KEY));
	controlMenu->AddItem(new BMenuItem("Stop", new BMessage(kStop), 'S', B_COMMAND_KEY));
	controlMenu->AddItem(new BMenuItem("Rewind", new BMessage(kRewind), 'R', B_COMMAND_KEY));
	controlMenu->AddSeparatorItem();
	controlMenu->AddItem(new BMenuItem("Step Forward", new BMessage(kForward), B_RIGHT_ARROW, B_COMMAND_KEY));
	controlMenu->AddItem(new BMenuItem("Step Back", new BMessage(kBack), B_LEFT_ARROW, B_COMMAND_KEY));
	controlMenu->AddSeparatorItem();
	controlMenu->AddItem(new BMenuItem("Loop", new BMessage(kLoop)));
	controlMenu->AddItem(new BMenuItem("Preserve Video Timing", new BMessage(kTiming)));
	fMenuBar->AddItem(controlMenu);
	AddChild(fMenuBar);
	
	fMenuBar->SubmenuAt(0)->ItemAt(0)->SetTarget(be_app);
	fMenuBar->SubmenuAt(0)->ItemAt(2)->SetTarget(be_app);
	
	fMenuBarHeight = fMenuBar->Frame().Height() + 1;

	fView = new BView(bounds, "VideoView", B_FOLLOW_ALL_SIDES, B_FRAME_EVENTS);
	fView->SetViewColor(0, 0, 0);
	AddChild(fView);
}

VideoWindow::~VideoWindow()
{
	PRINT(("VideoWindow::~VideoWindow()\n"));

	if (Lock()) {
		CloseFile();
		DeleteBuffer();
		Unlock();
	}
}

void VideoWindow::OpenFile(entry_ref *ref)
{
	PRINT(("VideoWindow::OpenFile()\n"));

	CloseFile();
	
	if (Lock()) {
		BEntry entry(ref);
		BPath path(ref);
		struct stat st;
		if (entry.Exists() && entry.GetStat(&st) == B_OK && S_ISREG(st.st_mode) &&
			(fFile = mpeg3_open(path.Path())) != NULL) {
			if (mpeg3_has_video(fFile)) {
				fVideoPlayer = new VideoPlayer(fFile, 0);
				fVideoPlayer->SetDrawBuffer(DrawBuffer, this);
			}
			if (mpeg3_has_audio(fFile)) {
				fAudioPlayer = new AudioPlayer(fFile, 0);
			}
			SetTitle(ref->name);
		}
		else {
			(new BAlert("Error", "Unable to play!", "Damn"))->Go();
		}
		Unlock();
	}
	
	SetRepeat(false);
}

void VideoWindow::CloseFile()
{
	PRINT(("VideoWindow::CloseFile()\n"));

	Stop();

	if (Lock()) {
		if (fFile != NULL) {
			delete fAudioPlayer;
			delete fVideoPlayer;

			fAudioPlayer = NULL;
			fVideoPlayer = NULL;
		
			mpeg3_close(fFile);
			fFile = NULL;
		}
		SetTitle("MPEG Player");
		Unlock();
	}
}

void VideoWindow::Start()
{
	PRINT(("VideoWindow::Start()\n"));

	if (fFile != NULL) {
		if (fVideoPlayer != NULL) {
			fVideoPlayer->Start();
			fView->SetViewColor(B_TRANSPARENT_32_BIT);
			fView->Invalidate();
		}
		if (fAudioPlayer != NULL)
			fAudioPlayer->Start();
	}
}

void VideoWindow::Stop()
{
	PRINT(("VideoWindow::Stop()\n"));

	if (fFile != NULL) {
		if (fVideoPlayer != NULL) {
			fVideoPlayer->Stop();
			fView->SetViewColor(0, 0, 0);
			fView->Invalidate();
		}
		if (fAudioPlayer != NULL)
			fAudioPlayer->Stop();
	}
}

void VideoWindow::SetScale(float scale)
{
	PRINT(("VideoWindow::SetScale()\n"));

	float width = 320, height = 0;
	
	if (fVideoPlayer != NULL) {
		width = fVideoPlayer->Width();
		height = fVideoPlayer->Height();
	}
	ResizeTo(scale * width - 1, scale * height + fMenuBar->Bounds().Height());
}

bool VideoWindow::Repeat()
{
	if (fVideoPlayer != NULL)
		return fVideoPlayer->Repeat();
	if (fAudioPlayer != NULL)
		return fAudioPlayer->Repeat();
	return false;
}
		
void VideoWindow::SetRepeat(bool repeat)
{
	PRINT(("VideoWindow::SetRepeat()\n"));

	fMenuBar->SubmenuAt(2)->ItemAt(7)->SetMarked(repeat);

	if (fVideoPlayer != NULL)
		fVideoPlayer->SetRepeat(repeat);
	if (fAudioPlayer != NULL)
		fAudioPlayer->SetRepeat(repeat);
}

bool VideoWindow::Full()
{
	return IsFullScreen();
}

void VideoWindow::SetFull(bool full)
{
	PRINT(("VideoWindow::SetFull()\n"));

	SetFullScreen(full);
	if (full)
		fMenuBar->ResizeTo(0, 0);
	else
		fMenuBar->ResizeToPreferred();
	
	if (full) {
		BRect screen(BScreen().Frame());
		Zoom(screen.LeftTop(), screen.Width(), screen.Height());
	}
	else {
		SetScale(1.0f);
	}
}

bigtime_t VideoWindow::CurrentTime()
{
	if (fVideoPlayer != NULL)
		return fVideoPlayer->CurrentTime();
	if (fAudioPlayer != NULL)
		return fAudioPlayer->CurrentTime();
	return 0;
}

void VideoWindow::SeekToTime(bigtime_t time)
{
	PRINT(("VideoWindow::SeekToTime()\n"));

	if (fVideoPlayer != NULL)
		fVideoPlayer->SeekToTime(time);
	if (fAudioPlayer != NULL)
		fAudioPlayer->SeekToTime(time);
}

void VideoWindow::MessageReceived(BMessage *message)
{
	PRINT(("VideoWindow::MessageReceived()\n"));

	switch (message->what) {
		case kScaleHalf:
			SetScale(0.5f);
			break;
			
		case kScaleNormal:
			SetScale(1.0f);
			break;
	
		case kScaleDouble:
			SetScale(2.0f);
			break;
	
		case kFullScreen:
			SetFull(!Full());
			break;
	
		case kPlay:
			Start();
			break;
		
		case kStop:
			Stop();
			break;
		
		case kRewind:
			SeekToTime(0);
			break;
			
		case kForward:
			// TODO
			break;
			
		case kBack:
			// TODO
			break;
			
		case kLoop:
			SetRepeat(!Repeat());
			break;
			
		case kTiming:
			// TODO
			break;
		
		case B_SIMPLE_DATA:
			{
				entry_ref ref;
				if (message->FindRef("refs", &ref) == B_OK) {
					BMessage refsMessage(B_REFS_RECEIVED);
					refsMessage.AddRef("refs", &ref);
					be_app->PostMessage(&refsMessage);
				}
			}
			break;
			
		default:
			BDirectWindow::MessageReceived(message);
			break;
	}
}
		
void VideoWindow::DirectConnected(direct_buffer_info *info)
{
	PRINT(("VideoWindow::DirectConnected()\n"));

	if (fLocker.Lock()) {
		switch (info->buffer_state & B_DIRECT_MODE_MASK) {
			case B_DIRECT_START:
				fConnected = true;
			case B_DIRECT_MODIFY:
				{
				fDirty = true;
				free(fInfo);
				int size = sizeof(direct_buffer_info) + info->clip_list_count * sizeof(clipping_rect);
				fInfo = (direct_buffer_info *) malloc(size);
				memcpy(fInfo, info, size);
				}
				break;
			case B_DIRECT_STOP:
				fConnected = false;
				break;
		}
		fLocker.Unlock();
	}
}


void VideoWindow::CreateBuffer(direct_buffer_info *info, int width, int height, float menuBarHeight)
{
	PRINT(("VideoWindow::CreateBuffer()\n"));
	
	if (info == NULL)
		return;
	
	if (fLocker.Lock()) {
		delete [] fBufferRows;
		delete [] fBufferClips;

		fBufferPixelFormat = info->pixel_format;
		
		fBufferNumClips = info->clip_list_count;
		fBufferClips = new blit_rect[fBufferNumClips];

		clipping_rect window = info->window_bounds;
		
		window.top += int(menuBarHeight);

		float hscale = float(width) / (window.right - window.left + 1);
		float vscale = float(height) / (window.bottom - window.top + 1);

		PRINT(("video=%d x %d, window=%ld x %ld, scale=%g,%g\n",
			width, height,
			window.right - window.left + 1, window.bottom - window.top + 1,
			hscale, vscale));
		
		int rows = 0;
		for (int index = 0; index < fBufferNumClips; index++) {
			clipping_rect clip = info->clip_list[index];
			
			if (clip.left < window.left)
				clip.left = window.left;
			if (clip.right > window.right)
				clip.right =  window.right;
			if (clip.top < window.top)
				clip.top = window.top;
			if (clip.bottom > window.bottom)
				clip.bottom = window.bottom;
			
			if (clip.left > clip.right || clip.top > clip.bottom) {
				clip.left = clip.right = window.right;
				clip.top = clip.bottom = window.bottom;
			}
						
			fBufferClips[index].src_left = int(hscale * (clip.left - window.left));
			fBufferClips[index].src_top = int(vscale * (clip.top - window.top));
			fBufferClips[index].src_width = int(hscale * (clip.right - clip.left + 1));
			fBufferClips[index].src_height = int(vscale * (clip.bottom - clip.top + 1));
			fBufferClips[index].dst_left = clip.left;			
			fBufferClips[index].dst_top = clip.top;
			fBufferClips[index].dst_width = clip.right - clip.left + 1;
			fBufferClips[index].dst_height = clip.bottom - clip.top + 1;

			rows += clip.bottom - clip.top + 1;
		}		
			
		fBufferRows = new unsigned char *[rows];
		for (int i = 0, index = 0; index < fBufferNumClips; index++) {
			blit_rect rect = fBufferClips[index];
			for (int row = 0; row < rect.dst_height; row++) {
				fBufferRows[i++] = (unsigned char *) info->bits +
							  rect.dst_left * ((info->bits_per_pixel + 7) / 8) +
							  (rect.dst_top + row) * info->bytes_per_row;
			}
		}

		fLocker.Unlock();
	}
}

void VideoWindow::DeleteBuffer()
{
	if (fLocker.Lock()) {
		delete [] fBufferRows;
		delete [] fBufferClips;
		
		fBufferRows = NULL;
		fBufferClips = NULL;
		
		fBufferPixelFormat = B_NO_COLOR_SPACE;
		fBufferNumClips = 0;

		fLocker.Unlock();
	}
}

void VideoWindow::Draw()
{
	//PRINT(("VideoWindow::Draw()\n"));

	if (fLocker.Lock()) {
		if (fVideoPlayer != NULL && fInfo != NULL &&
			(fDirty || fBufferRows == NULL || fBufferClips == NULL)) {
			CreateBuffer(fInfo, fVideoPlayer->Width(), fVideoPlayer->Height(),
				IsFullScreen() ? 0 : fMenuBarHeight);
		}
		if (fConnected && fVideoPlayer != NULL &&
				fBufferRows != NULL && fBufferClips != NULL) {
			if (fDirty || fVideoPlayer->CurrentFrame() < fVideoPlayer->CountFrames()) {
				fVideoPlayer->ReadFrame(fBufferRows, fBufferPixelFormat,
					fBufferClips, fBufferNumClips);
				fDirty = false;
			}
		}
		fLocker.Unlock();
	}
}

void VideoWindow::DrawBuffer(void *cookie)
{
	((VideoWindow *) cookie)->Draw();
}


////////////////////////////////////////////////////////////////////////////////

class VideoApp : public BApplication {
public:
	VideoApp();
	
	virtual ~VideoApp();
	
	virtual void MessageReceived(BMessage *message);
	
	virtual void RefsReceived(BMessage *message);
	
	virtual void ArgvReceived(int32 argc, char *argv[]);

	virtual void ReadyToRun();
	
	virtual void AboutRequested();
	
	void OpenFile(entry_ref *ref);

private:
	VideoWindow *fWindow;
	BFilePanel *fPanel;
};

VideoApp::VideoApp()
	:	BApplication("application/x-vnd.pel.MPEGPlayer"),
		fWindow(NULL),
		fPanel(NULL)
{
	PRINT(("VideoApp::VideoApp()\n"));

	fPanel = new BFilePanel(B_OPEN_PANEL);

	fWindow = new VideoWindow(BRect(100, 100, 100+320-1, 100+17+240-1),
		 "MPEG Player", B_TITLED_WINDOW, B_QUIT_ON_WINDOW_CLOSE);

	fWindow->Show();
	fWindow->Sync();
}

VideoApp::~VideoApp()
{
	PRINT(("VideoApp::~VideoApp()\n"));
}

void VideoApp::RefsReceived(BMessage *message)
{
	PRINT(("VideoApp::RefsReceived()\n"));

	entry_ref ref;
	if (message->FindRef("refs", &ref) == B_OK)
		OpenFile(&ref);
}

void VideoApp::ArgvReceived(int32 argc, char *argv[])
{
	PRINT(("VideoApp::ArgvReceived()\n"));

	if (argc == 2) {
		entry_ref ref;
		if (get_ref_for_path(argv[1], &ref) == B_OK)
			OpenFile(&ref);
	}
}	

void VideoApp::ReadyToRun()
{
	PRINT(("VideoApp::ReadyToRun()\n"));
}

void VideoApp::MessageReceived(BMessage *message)
{
	PRINT(("VideoApp::MessageReceived()\n"));
	
	switch (message->what) {
		case kOpen:
			fPanel->Show();
			break;
		
		default:
			BApplication::MessageReceived(message);
			break;
	}
}
	

void VideoApp::AboutRequested()
{
	PRINT(("VideoApp::AboutRequested()\n"));

	BAlert *alert = new BAlert("About",
		"MPEG Player\n"
		"by Carlos Hasan\n\n"
		"MPEG library written by Adam Williams",
		"Close", NULL, NULL, B_WIDTH_AS_USUAL, B_INFO_ALERT);
	alert->Go();
}

void VideoApp::OpenFile(entry_ref *ref)
{
	PRINT(("VideoApp::OpenFile(\"%s\")\n", ref->name));
	
	if (fWindow->Lock()) {
		fWindow->CloseFile();
		fWindow->Unlock();
	}
	
	if (fWindow->Lock()) {			
		fWindow->OpenFile(ref);
		fWindow->SetScale(1.0f);
		fWindow->Start();
		fWindow->Unlock();
	}
}

	
int main()
{
	VideoApp app;
	app.Run();
	return 0;
}